#
#  BLAST Decompressor  -  a script for decompressing the compressed BLAST output
#
#  Version 0.2.2 (June 27, 2012)
#
#  Copyright (c) 2012 Kirill Kryukov
#
#  This software is provided 'as-is', without any express or implied
#  warranty. In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#


#
# This decompressor is just a proof of concept. It does not aim to reproduce
# the original BLAST output.
#


use strict;
use File::Glob ':glob';



my %code_to_frame = (
    '+' => 'Plus/Plus',
    '-' => 'Plus/Minus',
    '1' => '+1',
    '2' => '+2',
    '3' => '+3',
    '4' => '-1',
    '5' => '-2',
    '6' => '-3',
    'A' => '+1/+1',
    'B' => '+1/+2',
    'C' => '+1/+3',
    'D' => '+1/-1',
    'E' => '+1/-2',
    'F' => '+1/-3',
    'G' => '+2/+1',
    'H' => '+2/+2',
    'I' => '+2/+3',
    'J' => '+2/-1',
    'K' => '+2/-2',
    'L' => '+2/-3',
    'M' => '+3/+1',
    'N' => '+3/+2',
    'O' => '+3/+3',
    'P' => '+3/-1',
    'Q' => '+3/-2',
    'R' => '+3/-3',
    'a' => '-1/+1',
    'b' => '-1/+2',
    'c' => '-1/+3',
    'd' => '-1/-1',
    'e' => '-1/-2',
    'f' => '-1/-3',
    'g' => '-2/+1',
    'h' => '-2/+2',
    'i' => '-2/+3',
    'j' => '-2/-1',
    'k' => '-2/-2',
    'l' => '-2/-3',
    'm' => '-3/+1',
    'n' => '-3/+2',
    'o' => '-3/+3',
    'p' => '-3/-1',
    'q' => '-3/-2',
    'r' => '-3/-3'
);



my @number_to_alphanumeric_code = (
    '0','1','2','3','4','5','6','7','8','9',
    'A','B','C','D','E','F','G','H','I','J',
    'K','L','M','N','O','P','Q','R','S','T',
    'U','V','W','X','Y','Z','a','b','c','d',
    'e','f','g','h','i','j','k','l','m','n',
    'o','p','q','r','s','t','u','v','w','x',
    'y','z'
);

my @alphanumeric_code_to_number = (-1) x 256;

my @number_to_symbol_code = (
    '!','"','#','$','%','&',"'",'(',')','*',
    '+',',','-','.','/',':',';','<','=','>',
    '?','@','[','\\',']','^','_','`','{','|',
    '}','~'
);

my @symbol_code_to_number = (-1) x 256;



my %genetic_code = (
    'AAA' => 'K',
    'AAC' => 'N',
    'AAG' => 'K',
    'AAT' => 'N',
    'ACA' => 'T',
    'ACC' => 'T',
    'ACG' => 'T',
    'ACT' => 'T',
    'AGA' => 'R',
    'AGC' => 'S',
    'AGG' => 'R',
    'AGT' => 'S',
    'ATA' => 'I',
    'ATC' => 'I',
    'ATG' => 'M',
    'ATT' => 'I',

    'CAA' => 'Q',
    'CAC' => 'H',
    'CAG' => 'Q',
    'CAT' => 'H',
    'CCA' => 'P',
    'CCC' => 'P',
    'CCG' => 'P',
    'CCT' => 'P',
    'CGA' => 'R',
    'CGC' => 'R',
    'CGG' => 'R',
    'CGT' => 'R',
    'CTA' => 'L',
    'CTC' => 'L',
    'CTG' => 'L',
    'CTT' => 'L',

    'GAA' => 'E',
    'GAC' => 'D',
    'GAG' => 'E',
    'GAT' => 'D',
    'GCA' => 'A',
    'GCC' => 'A',
    'GCG' => 'A',
    'GCT' => 'A',
    'GGA' => 'G',
    'GGC' => 'G',
    'GGG' => 'G',
    'GGT' => 'G',
    'GTA' => 'V',
    'GTC' => 'V',
    'GTG' => 'V',
    'GTT' => 'V',

    'TAA' => '*',
    'TAC' => 'Y',
    'TAG' => '*',
    'TAT' => 'Y',
    'TCA' => 'S',
    'TCC' => 'S',
    'TCG' => 'S',
    'TCT' => 'S',
    'TGA' => '*',
    'TGC' => 'C',
    'TGG' => 'W',
    'TGT' => 'C',
    'TTA' => 'L',
    'TTC' => 'F',
    'TTG' => 'L',
    'TTT' => 'F'
);



my %aa_positives = (
    'AS' => 1,
    'SA' => 1,

    'BD' => 1,
    'DB' => 1,
    'BE' => 1,
    'EB' => 1,
    'BN' => 1,
    'NB' => 1,
    'BZ' => 1,
    'ZB' => 1,

    'DE' => 1,
    'ED' => 1,
    'DN' => 1,
    'ND' => 1,
    'DZ' => 1,
    'ZD' => 1,

    'EK' => 1,
    'KE' => 1,
    'EQ' => 1,
    'QE' => 1,
    'EZ' => 1,
    'ZE' => 1,

    'FW' => 1,
    'WF' => 1,
    'FY' => 1,
    'YF' => 1,

    'HN' => 1,
    'NH' => 1,
    'HY' => 1,
    'YH' => 1,

    'IL' => 1,
    'LI' => 1,
    'IM' => 1,
    'MI' => 1,
    'IV' => 1,
    'VI' => 1,

    'KQ' => 1,
    'QK' => 1,
    'KR' => 1,
    'RK' => 1,
    'KZ' => 1,
    'ZK' => 1,

    'LM' => 1,
    'ML' => 1,
    'LV' => 1,
    'VL' => 1,

    'MV' => 1,
    'VM' => 1,

    'NS' => 1,
    'SN' => 1,

    'QR' => 1,
    'RQ' => 1,
    'QZ' => 1,
    'ZQ' => 1,

    'ST' => 1,
    'TS' => 1,

    'WY' => 1,
    'YW' => 1
);



my ($blastdbcmd_path,$q_path,$db_path,$db_names_path,$db_lengths_path,$infile_mask,$outext) = ('blastdbcmd','','','','','','blastdecompressed');

my $argc = scalar(@ARGV);
if (!$argc)
{
    print STDERR "Usage: perl blast_decompressor.pl options [<input >output]\n",
                 "Options:\n",
                 "    -query <query file>\n",
                 "    -db <database path>\n",
                 "    -dbnames <file with database sequence names>\n",
                 "    -dblengths <file with database sequence lengths>\n",
                 "    -blastdbcmd <path to blastdbcmd>\n",
                 "    -in <mask of files to decompress>\n",
                 "    -outext <output file extension>\n";
    exit;
}

for (my $i=0; $i<$argc; $i++)
{
    if ($i < $argc-1)
    {
        if ($ARGV[$i] eq '-query') { $i++; $q_path = $ARGV[$i]; }
        if ($ARGV[$i] eq '-db') { $i++; $db_path = $ARGV[$i]; }
        if ($ARGV[$i] eq '-dbnames') { $i++; $db_names_path = $ARGV[$i]; }
        if ($ARGV[$i] eq '-dblengths') { $i++; $db_lengths_path = $ARGV[$i]; }
        if ($ARGV[$i] eq '-blastdbcmd') { $i++; $blastdbcmd_path = $ARGV[$i]; }
        if ($ARGV[$i] eq '-in') { $i++; $infile_mask = $ARGV[$i]; }
        if ($ARGV[$i] eq '-outext') { $i++; $outext = $ARGV[$i]; }
    }
}



for (my $i=0; $i<62; $i++)
{
    $alphanumeric_code_to_number[ord($number_to_alphanumeric_code[$i])] = $i;
}
for (my $i=0; $i<32; $i++)
{
    $symbol_code_to_number[ord($number_to_symbol_code[$i])] = $i;
}



#
# Loading query sequences.
#
if (!$q_path) { die "Query file is not specified\n"; }
my %q_name_to_index = ();
my @q_index_to_name = ();
my @qseq = ();
my $n_qseq = 0;
{
    open(my $Q,'<',$q_path) or die "Can't open query file \"$q_path\"\n";
    while (<$Q>)
    {
        chomp;
        if (/^>(.+)$/)
        {
            my $name = $1;

            if ($q_name_to_index{$name}) { die "Duplicate name: $name\n"; }
            $n_qseq++;
            $q_name_to_index{$name} = $n_qseq;
            $q_index_to_name[$n_qseq] = $name;
            next;
        }
        if (!$n_qseq) { next; }
        $qseq[$n_qseq] .= $_;
    }
    close $Q;
}
print STDERR "Loaded $n_qseq query sequences\n";



#
# Loading database sequence names.
#
my %s_name_to_index = ();
my @s_index_to_name = ();
my $n_s_names = 0;
{
    if ($db_names_path eq '') { $db_names_path = "$db_path.names"; }
    if (!-e $db_names_path) { die "Can't find the database sequence names file \"$db_names_path\"\n"; }
    open (my $N,'<',$db_names_path) or die "Can't open database names file \"$db_names_path\"\n";
    while (<$N>)
    {
        chomp;
        if (!$_) { next; }
        #$_ =~ s/^>//;
        $_ =~ s/\s->\s.*$//;
        if (!$_) { die "Empty name found, #" . ($n_s_names+1) . "\n"; }
        my $name = $_;
        #if ($name =~ /^(gi\|\d+)/) { $name = $1; }
        if ($s_name_to_index{$name}) { die "Duplicate name: $name\n"; }
        $n_s_names++;
        $s_name_to_index{$name} = $n_s_names;
        $s_index_to_name[$n_s_names] = $name;
    }
    close $N;
}
#print STDERR "Loaded $n_s_names (", scalar(@s_index_to_name), ") database sequence names\n";
print STDERR "Loaded $n_s_names database sequence names\n";



#
# Loading database sequence lengths.
#
my @s_index_to_seq_length = (0);
{
    if ($db_lengths_path eq '') { $db_lengths_path = "$db_path.seq-length"; }
    if (!-e $db_lengths_path) { die "Can't find the database sequence lengths file \"$db_lengths_path\"\n"; }
    open (my $N,'<',$db_lengths_path) or die "Can't open database lengths file \"$db_lengths_path\"\n";
    while (<$N>)
    {
        chomp;
        if (!/^\d+$/) { die "Can't parse sequence length \"$_\"\n"; }
        push @s_index_to_seq_length, $_;
    }
    close $N;
}
print STDERR "Loaded ", scalar(@s_index_to_seq_length)-1, " database sequence lengths\n";
if (scalar(@s_index_to_seq_length) != scalar(@s_index_to_name)) { die "Number of lengths differs from number of names\n"; }



my @infiles = ();
if ($infile_mask ne '') { @infiles = bsd_glob($infile_mask); }
my $n_infiles = scalar @infiles;
if ($n_infiles)
{
    foreach my $infile (sort @infiles)
    {
        my $outfile = $infile;
        $outfile =~ s/\.[^\.]+$//;
        $outfile .= '.' . $outext;
        open (my $IN,'<',$infile) or die "Can't open file \"$infile\"\n";
        open (my $OUT,'>',$outfile) or die "Can't create file \"$outfile\"\n";
        print STDERR "Decompressing $infile -> $outfile\n";
        decompress_stream($IN,$OUT);
        close $IN;
        close $OUT;
    }
}
else
{
    my $IN = *STDIN;
    my $OUT = *STDOUT;
    print STDERR "Decompressing stdin -> stdout\n";
    decompress_stream($IN,$OUT);
}



sub decompress_stream
{
    my ($IN,$OUT) = @_;
    binmode $OUT;
    my ($qi,$si,$nhits_total,$nhits) = (0,0,0,0);
    while (<$IN>)
    {
        if ($_ eq '') { next; }
        if ($_ =~ /^#/) { next; }
        chomp;

        if (/^\s{2}/)
        {
            if (!$nhits and $qi) { print $OUT "\n\n***** No hits found *****\n"; }
            $nhits_total += $nhits;
            $nhits = 0;
            if (!/^\s{2}(\S+)$/) { die "Can't parse query index \"$_\"\n"; }
            my $qi_code = $1;
            $qi = decode_integer_base94($qi_code);
            if ($qi < 1 or $qi > $n_qseq) { die "Unknown query sequence index: \"$qi_code\" = $qi\n"; }
            my $s = 'Query= ' . $q_index_to_name[$qi] . ' ';
            $s =~ s/(.{1,79})\s/$1\n/g;
            print $OUT "\n\n$s\nLength=", length($qseq[$qi]), "\n";
            next;
        }

        if (/^\s/)
        {
            if (!/^\s(\S+)$/) { die "Can't parse database sequence index \"$_\"\n"; }
            my $si_code = $1;
            $si = decode_integer_base94($si_code);
            if ($si < 1 or $si > $n_s_names) { die "Unknown database sequence index: \"$si_code\" = $si\n"; }

            my $name = $s_index_to_name[$si];
            if ($name =~ /^>/)
            {
                my @names = split '>', substr($name,1);
                my $s = $names[0];
                foreach (my $i=1; $i<scalar(@names); $i++) { $s .= "\n " . $names[$i]; }
                $name = $s;
            }
            else
            {
                $name =~ s/ >/ /g;
                $name =~ s/>/ /g;
            }
            print $OUT "\n\n>$name\n";
            print $OUT "Length=", $s_index_to_seq_length[$si], "\n";
            next;
        }

        if (!/^(\S)(\S+)\s(\S+)\s(\S*)\s(\S*)\s(\S+)\s(\S+)\s(\S*)\s(\S+)\s(\S+)(.*)$/) { die "Can't parse hit \"$_\"\n"; }
        my ($frame_code,$aln_length_code,$identities_code,$positives_code,$ngaps_code,$score_code,$e_code,$nparts_code,$qstart_code,$sstart_code,$qsgaps) = ($1,$2,$3,$4,$5,$6,$7,$8,$9,$10,$11);

        if (!exists $code_to_frame{$frame_code}) { die "Unknown frame code: \"$frame_code\"\n"; }
        my $frame_str = $code_to_frame{$frame_code};
        my ($q_strand_minus,$s_strand_minus,$translated_query,$translated_subject) = (0,0,0,0);
        if ($frame_code eq '-' or $frame_code eq '+')
        {
            if ($frame_code eq '-') { $s_strand_minus = 1; }
        }
        elsif (ord($frame_code) >= ord('1') and ord($frame_code) <= ord('6'))
        {
            if (ord($frame_code) >= ord('4')) { $q_strand_minus = 1; }
            $translated_query = 1;
        }
        elsif ( (ord($frame_code) >= ord('A') and ord($frame_code) <= ord('Z')) or 
                (ord($frame_code) >= ord('a') and ord($frame_code) <= ord('z')) )
        {
            $frame_str =~ /^([\+\-])\d+\/([\+\-])\d+$/ or die;
            if ($1 eq '-') { $q_strand_minus = 1; }
            if ($2 eq '-') { $s_strand_minus = 1; }
            $translated_query = 1;
            $translated_subject = 1;
        }
        my $amino_acid_alignment = ($translated_query || $translated_subject) ? 1 : 0;

        my $aln_length = decode_integer_base94($aln_length_code);
        my $identities = decode_integer_base94($identities_code);
        my $positives = ($positives_code eq '') ? -1 : decode_integer_base94($positives_code);
        my $ngaps = ($ngaps_code eq '') ? 0 : decode_integer_base94($ngaps_code);
        my $score = decode_scientific($score_code);
        my $e = decode_scientific($e_code);
        my $nparts = ($nparts_code eq '') ? 1 : decode_integer_alphanumeric($nparts_code);
        my $qstart = decode_integer_base94($qstart_code);
        my $sstart = decode_integer_base94($sstart_code);

        if ($ngaps > 0 and $qsgaps eq '') { die "Error: $ngaps gaps, but empty gap pattern string, hit: \"$_\"\n"; }
        if ($ngaps == 0 and $qsgaps ne '') { die "Error: 0 gaps, but non-empty gap pattern string, hit: \"$_\"\n"; }
        my ($qngaps,$sngaps,$qlen,$slen) = (0,0,-1,-1);
        my @qint = ();
        my @sint = ();

        if ($qsgaps eq '')
        {
            $qlen = $aln_length;
            $slen = $aln_length;
        }
        else
        {
            if ($qsgaps !~ /^\s(\S*)\s(\S*)$/) { die "Can't parse gap patterns, hit: \"$_\"\n"; }
            my ($q_gaps_code,$s_gaps_code) = ($1,$2);
            if ($q_gaps_code ne '') { parse_gap_pattern($q_gaps_code,\@qint,\$_); }
            if ($s_gaps_code ne '') { parse_gap_pattern($s_gaps_code,\@sint,\$_); }
            for (my $i=1; $i<scalar(@qint); $i+=2) { $qngaps += $qint[$i]; }
            for (my $i=1; $i<scalar(@sint); $i+=2) { $sngaps += $sint[$i]; }
            $qlen = $aln_length - $qngaps;
            $slen = $aln_length - $sngaps;
            my $qsum = 0; for (my $i=0; $i<scalar(@qint); $i++) { $qsum += $qint[$i]; }
            my $ssum = 0; for (my $i=0; $i<scalar(@sint); $i++) { $ssum += $sint[$i]; }
            push @qint, $aln_length-$qsum;
            push @sint, $aln_length-$ssum;
        }

        if ($translated_query) { $qlen *= 3; }
        if ($translated_subject) { $slen *= 3; }

        my $qnint = scalar @qint;
        my $snint = scalar @sint;
        my $qend = $qstart + $qlen - 1;
        my $send = $sstart + $slen - 1;

        my $qs = substr(@qseq[$qi],$qstart-1,$qlen);
        if ($q_strand_minus) { $qs  = reverse($qs); $qs =~ tr/atcgATCG/tagcTAGC/; }
        if ($translated_query) { $qs = translate($qs); }

        my $s_strand_str = $s_strand_minus ? 'minus' : 'plus';
        my $s_accession = $s_index_to_name[$si];
        $s_accession =~ s/\s.*$//;
        $s_accession =~ s/^>//;
        my $cmd = $blastdbcmd_path .
                  ' -db ' . $db_path .
                  ' -entry "' . $s_accession . '"' .
                  ' -range ' . $sstart . '-' . ($sstart+$slen-1) .
                  ' -strand ' . $s_strand_str .
                  ' -outfmt %s';
        my $sseq = `$cmd`;
        chomp $sseq;
        if ($sseq =~ /^(.+?)[\r\n]/) { $sseq = $1; }
        if ($translated_subject) { $sseq = translate($sseq); }

        my $qseq_g = '';
        if ($qnint)
        {
            my ($pos,$gap) = (0,0);
            for (my $ii=0; $ii<$qnint; $ii++)
            {
                if ($gap) { $qseq_g .= ('-') x $qint[$ii]; } 
                else { $qseq_g .= substr($qs,$pos,$qint[$ii]); $pos += $qint[$ii]; }
                $gap = 1 - $gap;
            }
        }
        else { $qseq_g = $qs; }

        my $sseq_g = '';
        if ($snint)
        {
            my ($pos,$gap) = (0,0);
            for (my $ii=0; $ii<$snint; $ii++)
            {
                if ($gap) { $sseq_g .= ('-') x $sint[$ii]; } 
                else { $sseq_g .= substr($sseq,$pos,$sint[$ii]); $pos += $sint[$ii]; }
                $gap = 1 - $gap;
            }
        }
        else { $sseq_g = $sseq; }

        my $id_line = '';
        if ($amino_acid_alignment)
        {
            for (my $pos=0; $pos<$aln_length; $pos++)
            {
                my $q = uc(substr($qseq_g,$pos,1));
                my $s = uc(substr($sseq_g,$pos,1));
                if ($q eq '-' or $s eq '-' or $q eq 'X' or $s eq 'X') { $id_line .= ' '; }
                elsif ($q eq $s) { $id_line .= $q; }
                elsif (exists $aa_positives{$q.$s}) { $id_line .= '+'; }
                else { $id_line .= ' '; }
            }
        }
        else
        {
            for (my $pos=0; $pos<$aln_length; $pos++)
            {
                my $q = uc(substr($qseq_g,$pos,1));
                my $s = uc(substr($sseq_g,$pos,1));
                if (($q eq 'A' or $q eq 'T' or $q eq 'C' or $q eq 'G') and $q eq $s) { $id_line .= '|'; } else { $id_line .= ' '; }
            }
        }

        my $qstart_str = $q_strand_minus ? $qend : $qstart;
        my $sstart_str = $s_strand_minus ? $send : $sstart;
        my $qend_str = $q_strand_minus ? $qstart : $qend;
        my $send_str = $s_strand_minus ? $sstart : $send;

        my $qstart_len = length($qstart_str);
        my $sstart_len = length($sstart_str);
        my $start_len = ($qstart_len > $sstart_len) ? $qstart_len : $sstart_len;

        print $OUT "\n Score = ", $score, ' bits, Expect';
        if ($nparts > 1) { print $OUT '(', $nparts, ')'; }
        print $OUT ' = ', $e, "\n";
        print $OUT ' Identities = ', $identities, '/' , $aln_length, ' (', sprintf('%.0f',$identities*100/$aln_length), '%)';
        if ($positives >= 0) { print $OUT ', Positives = ', $positives, '/', $aln_length, ' (', sprintf('%.0f',$positives*100/$aln_length), '%)'; }
        print $OUT ', Gaps = ', $ngaps, '/', $aln_length, ' (', sprintf('%.0f',$ngaps*100/$aln_length), '%)', "\n";
        print $OUT ' ', (($translated_query or $translated_subject) ? 'Frame' : 'Strand'), ' = ', $frame_str, "\n\n";
        print $OUT sprintf('Query  %-*d  ',$start_len,$qstart_str), $qseq_g, '  ', $qend_str, "\n";
        print $OUT (' ' x ($start_len + 9)), $id_line, "\n";
        print $OUT sprintf('Sbjct  %-*d  ',$start_len,$sstart_str), $sseq_g, '  ', $send_str, "\n";

        $nhits++;
    }
    if (!$nhits) { print $OUT "\n\n***** No hits found *****\n"; }
    $nhits_total += $nhits;
    print STDERR "    $nhits_total hits\n";
}



sub translate
{
    my ($nucseq) = @_;
    $nucseq =~ tr/uUrRyYsSwWkKmMbBdDhHvV/tTnNnNnNnNnNnNnNnNnNnN/;
    my $pseq = '';
    my $nuclen = length($nucseq) - (length($nucseq)%3);
    for (my $i=0; $i<$nuclen; $i+=3)
    {
        my $codon = uc(substr($nucseq,$i,3));
        if ($codon =~ /[nN]/) { $pseq .= 'X'; }
        elsif (exists $genetic_code{$codon}) { $pseq .= $genetic_code{$codon}; }
        else { die "Unknown translation for codon \"$codon\"\n"; }
    }
    return $pseq;
}



sub decode_integer_base94
{
    my ($s) = @_;
    if ($s eq '') { return undef; }
    my $n = 0;
    my $l = length($s);
    for (my $i=0; $i<$l; $i++)
    {
        $n = $n * 94 + ord(substr($s,$i,1)) - 33;
    }
    return $n;
}



sub decode_integer_alphanumeric
{
    my ($s) = @_;
    if ($s eq '') { return undef; }
    my $n = 0;
    my $l = length($s);
    for (my $i=0; $i<$l; $i++)
    {
        $n = $n * 62 + $alphanumeric_code_to_number[ord(substr($s,$i,1))];
    }
    return $n;
}



sub decode_integer_symbol
{
    my ($s) = @_;
    if ($s eq '') { return undef; }
    my $n = 0;
    my $l = length($s);
    for (my $i=0; $i<$l; $i++)
    {
        $n = $n * 32 + $symbol_code_to_number[ord(substr($s,$i,1))];
    }
    return $n;
}



sub decode_float
{
    my ($code) = @_;
    my $di = index($code,'.');
    if ($di >= 0)
    {
        return decode_integer_alphanumeric(substr($code,0,$di)) . '.' . reverse(decode_integer_alphanumeric(substr($code,$di+1)));
    }
    else
    {
        return decode_integer_alphanumeric($code);
    }
}



sub decode_scientific
{
    my ($code) = @_;
    my $pi = index($code,'+');
    my $mi = index($code,'-');
    if ($pi >= 0 or $mi >= 0)
    {
        my $di = ($pi >= 0 and $mi >= 0) ? (($pi < $mi) ? $pi : $mi) : (($pi >= 0) ? $pi : $mi);
        return decode_float(substr($code,0,$di)) . 'e' . substr($code,$di,1) . sprintf("%03d",decode_integer_base94(substr($code,$di+1)));
    }
    else
    {
        return decode_float($code);
    }
}



sub parse_gap_pattern
{
    my ($code,$int,$hit) = @_;
    if ($alphanumeric_code_to_number[ord(substr($code,0,1))] < 0) { die "Can't parse gap pattern \"$code\" in hit \"$$hit\"\n"; }
    my $pe = -1;
    my $s = '';
    my $a = 1;
    my $l = length($code);
    for (my $i=0; $i<$l; $i++)
    {
        my $c = substr($code,$i,1);
        my $ca = ($alphanumeric_code_to_number[ord($c)] >= 0) ? 1 : 0;
        if ($ca == $a) { $s .= $c; next; }
        push @{$int}, ($a ? decode_integer_alphanumeric($s) : decode_integer_symbol($s));
        $s = $c;
        $a = $ca;
    }
    if ($a) { die "Last part of the gap pattern \"$code\" encodes non-gaps in hit \"$$hit\"\n"; }
    push @{$int}, decode_integer_symbol($s);
}
